/**
 * @file gpb.hpp
 * @date 2016-08-03
 */

// Copyright (C) 2016 Usagi
// Licensed under the MIT License.

// for Siv3D March 2016

#pragma once

#include <Siv3D.hpp>

#define BUFBYTE (4096)
#define LARGE_ABS (9999999.0)

typedef unsigned char u8;

#define GTYPE_SCENE (1)
#define GTYPE_NODE (2)
#define GTYPE_ANIMATIONS (3)
#define GTYPE_MESH (0x22)


/**
 * X[p[NX
 */
struct GpbChunk {
	/// <summary></summary>
	String name;
	/// <summary>^Cv</summary>
	int type;
	/// <summary>t@C̎f[^ʒu</summary>
	int offsetTop;

	/// <summary>wb_[̈̕ʒu</summary>
	int offsetWrite;
};

	/// <summary>_</summary>
	struct GpbVertex {
		/// <summary>ʒu</summary>
		Vec3 position;
		/// <summary>@</summary>
		Vec3 normal;
		/// <summary>@</summary>
		Vec2 texcoord;

		/// <summary>TEXCOORD1(gp)</summary>
		//Vec4 texcoord1;

#if USESKIN != 0
		/// <summary>EFCg</summary>
		Vec4 blendWeights = { 1,0,0,0 };
		/// <summary>CfbNX</summary>
		Vec4 blendIndices = { 0,0,0,0 };
#endif
	};


	/// <summary>ʒ_</summary>
	struct GpbPart {
		/// <summary>GL_TRIANGLESȂ</summary>
		int prim = 4;
		/// <summary>unsigned short</summary>
		int format = 0x00001403;

		/// <summary>ʒ_</summary>
		Array<int> faceIndices;
	};

	/// <summary>bVCu</summary>
	struct GpbMesh : public GpbChunk {
		/// <summary>lfloat</summary>
		Array<Point> attrNums;

		/// <summary>_Q</summary>
		Array<GpbVertex> vs;

		/// <summary>oEfBO`ŏl</summary>
		Vec3 mini;
		/// <summary>oEfBO</summary>
		Vec3 maxi;
		/// <summary>ES</summary>
		Vec3 center;
		/// <summary>Ea</summary>
		double radius;
			

		/// <summary>(L)</summary>
		Array<GpbPart> parts;
	};

	class GpbNode;

	/// <summary>m[h܂̓WCg</summary>
	class GpbNode: public GpbChunk {
	public:
		/// <summary>1:node, 2:joint</summary>
		int nj = 1;

		/// <summary>s</summary>
		Mat4x4 m;

		/// <summary>em[h</summary>
		String parent = L"";

		/// <summary>q`NBdĂƂ</summary>
		Array<String> children;

		/// <summary>u8 0ŒƂ</summary>
		int isCam = 0;
		/// <summary>u8 0ŒƂ</summary>
		int isLgt = 0;


		/// <summary>QƖ #ȂƂ</summary>
		String ref = L"";
		/// <summary>u8 skin 0ŒƂ</summary>
		int isSkin = 0;
		/// <summary>ގ</summary>
		Array<String> mtls;

		/// <summary>RXgN^</summary>
		GpbNode() {
			this->m = Mat4x4::Identity();
		};
	};

	/// <summary>ގ</summary>
	struct GpbMaterial {
		/// <summary>.materialɏo()</summary>
		String name;

		/// <summary>gU</summary>
		ColorF diffuse;

		/// <summary>XyL</summary>
		double power;
		/// <summary>XyLF</summary>
		ColorF specular;
		/// <summary></summary>
		ColorF ambient;

		/// <summary>֊sAe</summary>
		int edgeFlag = 0;

		/// <summary>ގɊ܂܂钸_</summary>
		int matVertexNum = 0;
		/// <summary>eNX`t@C</summary>
		String texName;
	};

	/// <summary>WCgB</summary>
	struct GpbJoint : public GpbNode {
	};

	/// <summary>L[t[</summary>
	struct GpbKey {
		/// <summary>^CX^v</summary>
		int timeStamp;
		/// <summary>]</summary>
		Vec4 q;
		/// <summary>sړ</summary>
		Vec3 v;
	};

	/// <summary>1̃^[Qbgɑ΂1</summary>
	struct GpbTrack {
		/// <summary>Ώ</summary>
		String target;
		/// <summary></summary>
		int attr = 16;

		/// <summary>L[t[</summary>
		Array<GpbKey> keys;


		/// <summary></summary>
		Array<int> ips;
	};

	/// <summary>1̃Aj[V</summary>
	struct GpbAnimation : public GpbChunk {
		/// <summary>q</summary>
		Array<GpbTrack> children;
	};

	/// <summary>Aj[Vi[</summary>
	struct GpbAnimations : public GpbChunk {
		/// <summary>q</summary>
		Array<GpbAnimation> children;
	};

	/// <summary>V[Lq</summary>
	struct GpbScene : public GpbNode {
		/// <summary>q`N</summary>
		Array<String> children;
		/// <summary></summary>
		String activeCam = L"";
		/// <summary></summary>
		ColorF amb = { 0,0,0 };
	};

	/// <summary>1̃t@C\</summary>
	struct GpbFile {
		/// <summary>f</summary>
		String name = L"model";
		/// <summary>Model name</summary>
		String nameE = L"model";


		/// <summary>eNX`</summary>
		Array<String> texNames;
		/// <summary>ގ</summary>
		Array<GpbMaterial> materials;


		/// <summary>bVCu̎̑</summary>
		Array<GpbMesh> meshes;

		/// <summary>WCg`N̎̑</summary>
		Array<GpbJoint> joints;

		/// <summary>bVQƂm[hŎ̑</summary>
		Array<GpbNode> refs;


		/// <summary>1̃V[</summary>
		GpbScene scene;
		/// <summary>1̃Aj[V(ɕ)</summary>
		GpbAnimations anim;

		/// <summary>zNA</summary>
		void clear() {
			this->joints.clear();
			this->refs.clear();

			this->scene.children.clear();
			this->anim.children.clear();

			this->meshes.clear();

			this->texNames.size();
			this->materials.size();
		}

		/// <summary>̂CfbNXT</summary>
		Point findIndexByName(String target) const {
			Point ret = { -1,-1 };

			for (size_t i = 0; i < this->refs.size(); ++i) {
				if (this->refs[i].name == target) {
					ret.x = 1;
					ret.y = i;
					break;
				}
			}

			if (ret.y < 0) {
				for (size_t i = 0; i < this->joints.size(); ++i) {
					if (this->joints[i].name == target) {
						ret.x = 2;
						ret.y = i;
						break;
					}
				}
			}
			return ret;
		};
	};

	/// <summary>t@C</summary>
	struct GpbFiler {
		/// <summary></summary>
		/// <param name="w">oCiC^[</param>
		/// <param name="s"></param>
		void ws(BinaryWriter& w, String s) const {
			char buf[BUFBYTE];

			//std::string ms = s.narrow(); // Shift_JIS
			std::string ms = ToUTF8(s); // UTF-8
			int len = ::strlen(ms.c_str());

			int size = Min(len, BUFBYTE - 4);
			::memcpy(buf, &size, 4);
			::memcpy(buf + 4, ms.c_str(), size);

			w.write(buf, (4 + size));
		}

		/// <summary>bVp[g</summary>
		void writePart(BinaryWriter& w, const GpbPart& p) const {
			w.write<int>(p.prim);
			w.write<int>(p.format);

			// _CfbNX
			auto faceVNum = p.faceIndices.size();
			w.write<int>(faceVNum * sizeof(unsigned short));
			for (const auto& idx : p.faceIndices) {
				w.write<unsigned short>(idx);
			}
		};

		/// <summary>bVCȕo</summary>
		void writeMesh(BinaryWriter& w, const GpbMesh& m) const {
			try {
				w.write<int>(m.attrNums.size());

				int attrNumSum = 0;
				for (const auto& un : m.attrNums) {
					w.write<int>(un.x);
					w.write<int>(un.y);
					attrNumSum += un.y;
				}

				int vNum = m.vs.size();
				w.write<int>(sizeof(float) * attrNumSum * vNum);
				for (const auto& v : m.vs) {
					w.write<float>(v.position.x);
					w.write<float>(v.position.y);
					w.write<float>(v.position.z);

					w.write<float>(v.normal.x);
					w.write<float>(v.normal.y);
					w.write<float>(v.normal.z);

					w.write<float>(v.texcoord.x);
					w.write<float>(v.texcoord.y);

#if USESKIN != 0
					w.write<float>(v.blendWeights.x);
					w.write<float>(v.blendWeights.y);
					w.write<float>(v.blendWeights.z);
					w.write<float>(v.blendWeights.w);

					w.write<float>(v.blendIndices.x);
					w.write<float>(v.blendIndices.y);
					w.write<float>(v.blendIndices.z);
					w.write<float>(v.blendIndices.w);
#endif
				}

				// ŏőȂ
				w.write<float>(m.mini.x);
				w.write<float>(m.mini.y);
				w.write<float>(m.mini.z);

				w.write<float>(m.maxi.x);
				w.write<float>(m.maxi.y);
				w.write<float>(m.maxi.z);

				w.write<float>(m.center.x);
				w.write<float>(m.center.y);
				w.write<float>(m.center.z);

				w.write<float>(s3d::Math::Sqrt(m.radius));


				////  mesh parts
				auto partNum = m.parts.size();
				w.write<int>(partNum);
				for (const auto& part : m.parts) {
					writePart(w, part);
				}

			}
			catch (...) {

			}
		};

		/// <summary>m[h`NoB
		/// q`NɂΉ邱ƁB</summary>
		/// <param name="n">m[h</param>
		void writeNode(BinaryWriter& w, GpbNode* node, GpbFile& f) const {
			try {
				node->offsetTop = w.getPos();

				GpbNode& n = *node;

				// 1:node, 2:joint
				w.write<int>(n.nj);

				// transform
				for (int i = 0; i < 4; ++i) {
					for (int j = 0; j < 4; ++j) {
						w.write<float>(n.m.r[i].m128_f32[j]);
					}
				}

				ws(w, n.parent);

				// children
				w.write<int>(n.children.size());
				for (auto child : n.children) {
					auto chunk = f.findIndexByName(child);
					if (chunk.y >= 0) {
						if (chunk.x == 1) {
							writeNode(w, &f.refs[chunk.y], f);
						}
						else {
							writeNode(w, &f.joints[chunk.y], f);
						}
					}
					else {
						Println(L"Not found", child);
					}
				}

				// camera
				w.write<u8>(n.isCam);
				// light
				w.write<u8>(n.isLgt);

				// f
				// bVQ
				ws(w, n.ref);
				if (n.ref.length > 0) {
					// XLtO
					w.write<u8>(n.isSkin);
					// ގJEg
					auto mtlNum = n.mtls.size();
					w.write<int>(mtlNum);
					for (const auto& mtl : n.mtls) {
						ws(w, mtl);
					}
				}
			}
			catch (...) {

			}
		};

		/// <summary>gbNo</summary>
		void writeTrack(BinaryWriter& w, const GpbTrack& anim) const {
			ws(w, anim.target);
			w.write<int>(anim.attr);

			int keyNum = anim.keys.size();
			w.write<int>(keyNum);
			for (int i = 0; i < keyNum; ++i) {
				w.write<int>(anim.keys[i].timeStamp);
			}

			w.write<int>(keyNum * 7);
			for (const auto& key : anim.keys) {
				w.write<float>(key.q.x);
				w.write<float>(key.q.y);
				w.write<float>(key.q.z);
				w.write<float>(key.q.w);

				w.write<float>(key.v.x);
				w.write<float>(key.v.y);
				w.write<float>(key.v.z);
			}

			int tanInNum = 0;
			w.write<int>(tanInNum);

			int tanOutNum = 0;
			w.write<int>(tanOutNum);

			auto ipNum = anim.ips.size();
			w.write<int>(ipNum);
			for (const auto& ip: anim.ips) {
				w.write<int>(ip);
			}
		};


		/// <summary>.gpbt@Co͂</summary>
		/// <param name="f">.gpb̏</param>
		/// <param name="w">oCiC^</param>
		bool SaveGpb(GpbFile& f, BinaryWriter& w) const {
			int isAnim = (f.anim.children.size() > 0) ? 1 : 0;

			//// wb_
			w.write<u8>(0xAB);
			w.write<u8>('G');
			w.write<u8>('P');
			w.write<u8>('B');
			w.write<u8>(0xBB);
			w.write<u8>(0x0D);
			w.write<u8>(0x0A);
			w.write<u8>(0x1A);
			w.write<u8>(0x0A);

			// ver
			w.write<u8>(0x01);
			w.write<u8>(0x02);


			// t@Xe[u
			int refNum = f.meshes.size() + f.joints.size() + f.refs.size() + 1 + isAnim;
			w.write<int>(refNum);
			for (auto& lib: f.meshes) {
				ws(w, lib.name);
				w.write<int>(lib.type);
				// ߂̂߂ɕۑĂ
				lib.offsetWrite = w.getPos();
				w.write<int>(0);
			}
			for (auto& j: f.joints) {
				ws(w, j.name);
				w.write<int>(j.type);
				j.offsetWrite = w.getPos();
				w.write<int>(0);
			}
			for (auto& m: f.refs) {
				ws(w, m.name);
				w.write<int>(m.type);
				m.offsetWrite = w.getPos();
				w.write<int>(0);
			}
			{
				ws(w, f.scene.name);
				w.write<int>(f.scene.type);
				f.scene.offsetWrite = w.getPos();
				w.write<int>(0);
			}
			if (isAnim) {
				ws(w, f.anim.name);
				w.write<int>(f.anim.type);
				f.anim.offsetWrite = w.getPos();
				w.write<int>(0);
			}


			//// 

			// bV
			int meshNum = f.meshes.size();
			w.write<int>(meshNum);
			for (auto& mesh : f.meshes) {
				mesh.offsetTop = w.getPos();
				writeMesh(w, mesh);
			}


			// IuWFNg
			int objNum = 1 + isAnim;
			w.write<int>(objNum);

			{ // V[
				f.scene.offsetTop = w.getPos();

				auto cnum = f.scene.children.size();
				w.write<int>(cnum);
				for (auto child: f.scene.children) {
					//child->offsetTop = w.getPos();

					auto chunk = f.findIndexByName(child);
					if (chunk.y >= 0) {
						if (chunk.x == 1) {
							writeNode(w, &f.refs[chunk.y], f);
						}
						else {
							writeNode(w, &f.joints[chunk.y], f);
						}
					}
					else {
						Println(L"Not found,", child);
					}
				}

				ws(w, f.scene.activeCam);
				w.write<float>(f.scene.amb.r);
				w.write<float>(f.scene.amb.g);
				w.write<float>(f.scene.amb.b);
			}


			//// Aj[V
			if (isAnim) {
				f.anim.offsetTop = w.getPos();

				auto aNum = f.anim.children.size();
				w.write<int>(aNum);
				for (auto& anim: f.anim.children) {
					//anim.offsetTop = w.getPos();
					ws(w, anim.name);

					auto tNum = anim.children.size();
					w.write<int>(tNum);
					for (const auto& track : anim.children) {
						writeTrack(w, track);
					}
				}
			}


			//// ߂ď
			for (const auto& lib : f.meshes) {
				w.setPos(lib.offsetWrite);
				w.write<int>(lib.offsetTop);
			}
			for (const auto& j : f.joints) {
				w.setPos(j.offsetWrite);
				w.write<int>(j.offsetTop);
			}
			for (const auto& m : f.refs) {
				w.setPos(m.offsetWrite);
				w.write<int>(m.offsetTop);
			}
			{
				w.setPos(f.scene.offsetWrite);
				w.write<int>(f.scene.offsetTop);
			}
			if (isAnim) {
				w.setPos(f.anim.offsetWrite);
				w.write<int>(f.anim.offsetTop);
			}

			w.close();
			return true;
		}

		/// <summary>}eAt@Co</summary>
		bool SaveMaterial(const GpbFile& f, TextWriter w) const {
			String templateFileName = L"template.material";
			String dir = FileSystem::ParentPath(FileSystem::ModulePath());

			TextReader tr(Format(dir,templateFileName),
				s3d::TextEncoding::UTF8);

			String whole;
			auto result = tr.readAll(whole);
			if (result == false) {
				Println(L"Not found template");
				w.close();
				return false;
			}
			w.writeln(whole);

			for (const auto& m : f.materials) {
				Array<String> ls;
				if (m.texName.isEmpty == false) {
					ls.push_back(Format(L"material ", m.name, L": Textured {"));

					ls.push_back(Format(L"\tu_specularExponent = ", m.power));

					ls.push_back(Format(PyFmt, L"\tu_ambientColor = {},{},{}",
						m.ambient.r, m.ambient.g, m.ambient.b));

					ls.push_back(L"\tsampler u_diffuseTexture {");
					ls.push_back(Format(L"\t\tpath = ", FileSystem::FileName(m.texName)));
					ls.push_back(L"\t}");
					ls.push_back(L"\ttechnique {");
					ls.push_back(L"\t\tpass {");
					ls.push_back(L"\t\t\tdefines = SPECULAR;MODULATE_ALPHA");
					ls.push_back(L"\t\t}");
					ls.push_back(L"\t}");
					ls.push_back(L"}");
				}
				else {
					ls.push_back(Format(L"material ", m.name, L": Colored {"));

					ls.push_back(Format(L"\tu_specularExponent = ", m.power));

					ls.push_back(Format(PyFmt, L"\tu_diffuseColor = {},{},{}, {}",
						m.diffuse.r, m.diffuse.g, m.diffuse.b, m.diffuse.a));
					ls.push_back(Format(PyFmt, L"\tu_ambientColor = {},{},{}",
						m.ambient.r, m.ambient.g, m.ambient.b));

					ls.push_back(L"\ttechnique {");
					ls.push_back(L"\t\tpass {");
					ls.push_back(L"\t\t\tdefines = SPECULAR;MODULATE_ALPHA");
					ls.push_back(L"\t\t}");
					ls.push_back(L"\t}");
					ls.push_back(L"}");
				}

				for (const auto& l : ls) {
					w.writeln(l);
				}
			}

			w.close();
			return true;
		};



		//// mFp

		/// <summary>t@C擪p[X</summary>
		/// <param name="f">ΉBt@Ci[</param>
		bool LoadGpb(BinaryReader& r, GpbFile& f) const {
			int v32;
			float vf;
			int isFont = 0;
			Array<GpbChunk> chunks;

			// wb_
			r.skip(11);

			int cNum = 0;
			r.read<int>(cNum);

			for (int i = 0; i < cNum; ++i) {
				GpbChunk chunk;

				chunk.name = rs(r);
				r.read<int>(chunk.type);
				r.read<int>(chunk.offsetTop);

				chunks.push_back(chunk);

				Println(PyFmt, L"{},{},{}",
					chunk.name, chunk.type, chunk.offsetTop);

				if (chunk.type == 128) {
					isFont = 1;
				}
			}

			if (isFont) {
				String name = rs(r);
				Println(L"font,", name);
				return true;
			}

			if (true) {
				//return true;
			}

			r.read<int>(v32);
			for (int i = 0; i < v32; ++i) {
				loadMesh(r, f);
			}


			r.read<int>(v32);
			for (int i = 0; i < v32; ++i) {
				auto cur = r.getPos();
				int type = -1;
				for (int j = 0; j < cNum; ++j) {
					if (chunks[j].offsetTop == cur) {
						type = chunks[j].type;
						break;
					}
				}

				if (type == 1) {
					int cNum = 0;
					r.read<int>(cNum);
					Println(L"scene child,", cNum);

					for (int j = 0; j < cNum; ++j) {
						loadNode(r, f);
					}

					rs(r);
					r.read<float>(vf);
					r.read<float>(vf);
					r.read<float>(vf);
				}
				else if (type == 2) {
					loadNode(r, f);
				}
				else if (type == 3) {
					loadAnim(r, f);
				}
				else {
					Println(L"Unknown object", L",", type);
					break;
				}
			}

			auto tail = r.getPos();
			Println(tail, L"/", r.size());


			
// e[ũWv̏ꍇ
			/*
			for (int i = 0; i < cNum; ++i) {
				r.setPos(chunks[i].offsetTop);
				
				switch (chunks[i].type) {
				case 1: // V[
					break;
				case 2: // m[h
					//LoadNode(r, f);
					break;
				case 3: // Aj[V
					loadAnim(r, f);
					break;
				case 32: // J
					break;
				case 33: // Cg
					break;
				case 34: // bV
					//LoadMesh(r, f);
					break;
				case 128: // tHg
					break;
				}
			}
			*/

			return true;
		};

		/// <summary>bV̓ǂݍ݁B</summary>
		bool loadMesh(BinaryReader& r, GpbFile& f) const {
			int v32 = 0;
			float vf = 0.0f;

			int num = 0;
			r.read<int>(num);

			int attrNumSum = 0;
			for(int i = 0; i < num; ++i) {
				r.read<int>(v32);
				r.read<int>(v32);
				attrNumSum += v32;
			}

			r.read<int>(num);
			int vNum = num / sizeof(float) / attrNumSum;
			for (int i = 0; i < vNum; ++i) {
				for (int j = 0; j < attrNumSum; ++j) {
					r.read<float>(vf);
				}
			}

			Vec3 mini = readF3(r);
			Vec3 maxi = readF3(r);
			Vec3 center = readF3(r);
			float radius;
			r.read<float>(radius);

			r.read<int>(num);
			for (int i = 0; i < num; ++i) {
				// tri
				r.read<int>(v32);
				// u16
				r.read<int>(v32);

				// oCg
				r.read<int>(v32);

				for (int i = 0; i < v32/2; ++i) {
					r.skip(2);
				}
			}


			Println(mini);
			Println(maxi);
			Println(center);
			Println(PyFmt, L"radius,{}", radius);

			return true;
		};

		/// <summary>J[VuɃm[hǂݎB</summary>
		bool loadNode(BinaryReader& r, GpbFile& f) const {
			try {
				int v32;
				u8 v8;
				float vf;

				auto top = r.getPos();

				GpbNode n;

				r.read<int>(n.nj);
				for (int i = 0; i < 4; ++i) {
					for (int j = 0; j < 4; ++j) {
						r.read<float>(vf);
						n.m.r[i].m128_f32[j] = vf;
					}
				}
				n.parent = rs(r);
				int cNum = 0;
				r.read<int>(cNum);
				Println(L"top,", top, L",c,", cNum);
				if (cNum > 0) {
					for (int i = 0; i < cNum; ++i) {
						loadNode(r, f);
					}
				}

				r.read<u8>(v8);
				n.isCam = v8;
				r.read<u8>(v8);
				n.isLgt = v8;

				n.ref = rs(r);
				if (n.ref.length > 0) {
					r.read<u8>(v8);
					n.isSkin = v8;
					if (n.isSkin) {
						Array<String> ss;

						for (int i = 0; i < 16; ++i) {
							r.read<float>(vf);
						}

						r.read<int>(v32);
						for (int i = 0; i < v32; ++i) {
							ss.push_back(rs(r));
						}
						r.read<int>(v32);
						for (int i = 0; i < v32 / 16; ++i) {
							r.skip(4 * 16);
						}
					}

					int mNum = 0;
					r.read<int>(mNum);
					for (int i = 0; i < mNum; ++i) {
						n.mtls.push_back(rs(r));
					}
				}

				if (true) {
					Println(n.nj, L",p,", n.parent, L",c,", cNum, L",", n.ref);
				}

			}
			catch (...) {
				return false;
			}

			return true;
		};

		/// <summary>^Cv3ǂ݂ƂB</summary>
		bool loadAnim(BinaryReader& r, GpbFile& f) const {
			try {
				int v32;
				float vf;

				GpbAnimations anims;

				int aNum = 0;
				r.read<int>(aNum);
				for (int i = 0; i < aNum; ++i) {
					GpbAnimation anim;
					anim.name = rs(r);

					int chNum = 0;
					r.read<int>(chNum);
					for (int j = 0; j < chNum; ++j) {
						GpbTrack ch;
						ch.target = rs(r);
						r.read<int>(ch.attr);
						int kNum = 0;
						r.read<int>(kNum);
						for (int k = 0; k < kNum; ++k) {
							GpbKey key;
							r.read<int>(key.timeStamp);
							ch.keys.push_back(key);
						}

						r.read<int>(v32);
						for (int l = 0; l < v32; ++l) {
							r.skip(4);
						}


						r.read<int>(v32);
						r.read<int>(v32);

						int ipNum = 0;
						r.read<int>(ipNum);
						for (int m = 0; m < ipNum; ++m) {
							r.read<int>(v32);
							ch.ips.push_back(v32);
						}

						anim.children.push_back(ch);
						//Println(L"ch,", ch.target, L",attr,", ch.attr, L",key,", ch.keys.size());
					}

					anims.children.push_back(anim);

					Println(L"anim,", anim.name, L",ch,", anim.children.size());
				}
			}
			catch (...) {
				return false;
			}
			return true;
		};

		/// <summary>^ǂݎ</summary>
		s3d::String rs(BinaryReader& r) const {
			char buf[BUFBYTE];
			int v32 = 0;
			r.read<int>(v32);

			v32 = s3d::Clamp<int>(v32, 0, BUFBYTE - 1);

			r.read(buf, v32);
			buf[v32] = 0;
			std::string s0(buf);
//			return Widen(s0);
			return FromUTF8(s0);
		};

		/// <summary>float3ǂݎ</summary>
		s3d::Vec3 readF3(BinaryReader& r) const {
			s3d::Vec3 ret;
			float vf;
			r.read<float>(vf);
			ret.x = vf;
			r.read<float>(vf);
			ret.y = vf;
			r.read<float>(vf);
			ret.z = vf;
			return ret;
		};
	};
